@@ -55,6 +55,7 @@
 
 static PyObject *sort_str, *reverse_str, *__setstate___str;
 static PyObject *_bucket_type_str, *max_internal_size_str, *max_leaf_size_str;
+static PyObject *__slotnames__str;
 static PyObject *ConflictError = NULL;
 
 static void PyVar_Assign(PyObject **v, PyObject *e) { Py_XDECREF(*v); *v=e;}
@@ -77,28 +78,126 @@ static void PyVar_Assign(PyObject **v, PyObject *e) { Py_XDECREF(*v); *v=e;}
 #error "PY_LONG_LONG required but not defined"
 #endif
 
+static int
+longlong_handle_overflow(PY_LONG_LONG result, int overflow)
+{
+    if (overflow)
+    {
+        PyErr_Clear();
+        /* Python 3 tends to have an exception already set, Python 2 not so
+           much. We want to consistently raise a TypeError.
+        */
+        PyErr_SetString(PyExc_TypeError, "couldn't convert integer to C long long");
+        return 0;
+    }
+    else if (result == -1 && PyErr_Occurred())
+        /* An exception has already been raised. */
+        return 0;
+    return 1;
+}
+
+
 #ifdef NEED_LONG_LONG_KEYS
+
+#if defined(ZODB_UNSIGNED_VALUE_INTS) || defined(ZODB_UNSIGNED_KEY_INTS)
+static int
+ulonglong_check(PyObject *ob)
+{
+#ifndef PY3K
+    if (PyInt_Check(ob))
+    {
+        long tmp;
+        tmp = PyInt_AS_LONG(ob);
+        if (tmp < 0) {
+            PyErr_SetString(PyExc_TypeError, "unsigned value less than 0");
+            return 0;
+        }
+        return 1;
+    }
+#endif
+
+    if (!PyLong_Check(ob))
+    {
+        return 0;
+    }
+
+    if (PyLong_AsUnsignedLongLong(ob) == (unsigned long long)-1 && PyErr_Occurred())
+    {
+        return 0;
+    }
+    return 1;
+}
+#endif /* defined(ZODB_UNSIGNED_VALUE_INTS) || defined(ZODB_UNSIGNED_KEY_INTS) */
+
 static int
 longlong_check(PyObject *ob)
 {
-    if (INT_CHECK(ob))
+#ifndef PY3K
+    /* Python's small integers can always fit into a long long. */
+    if (PyInt_Check(ob))
         return 1;
+#endif
 
     if (PyLong_Check(ob)) {
         int overflow;
-        (void)PyLong_AsLongLongAndOverflow(ob, &overflow);
-        if (overflow)
-            goto overflow;
-        return 1;
+        PY_LONG_LONG result;
+        result = PyLong_AsLongLongAndOverflow(ob, &overflow);
+        return longlong_handle_overflow(result, overflow);
     }
     return 0;
-overflow:
-    PyErr_SetString(PyExc_ValueError,
-                    "longlong_check: long integer out of range");
-    return 0;
 }
+
 #endif
 
+#if defined(ZODB_UNSIGNED_VALUE_INTS) || defined(ZODB_UNSIGNED_KEY_INTS)
+static PyObject *
+ulonglong_as_object(unsigned PY_LONG_LONG val)
+{
+    if ((val > LONG_MAX))
+        return PyLong_FromUnsignedLongLong(val);
+    return UINT_FROM_LONG((unsigned long)val);
+}
+
+static int
+ulonglong_convert(PyObject *ob, unsigned PY_LONG_LONG *value)
+{
+    unsigned PY_LONG_LONG val;
+
+#ifndef PY3K
+    if (PyInt_Check(ob))
+    {
+        long tmp;
+        tmp = PyInt_AS_LONG(ob);
+        if (tmp < 0) {
+            PyErr_SetString(PyExc_TypeError, "unsigned value less than 0");
+            return 0;
+        }
+        (*value) = (unsigned PY_LONG_LONG)tmp;
+        return 1;
+    }
+#endif
+
+    if (!PyLong_Check(ob))
+    {
+        PyErr_SetString(PyExc_TypeError, "expected integer key");
+        return 0;
+    }
+
+    val = PyLong_AsUnsignedLongLong(ob);
+    if (val == (unsigned long long)-1 && PyErr_Occurred())
+    {
+        if (PyErr_ExceptionMatches(PyExc_OverflowError))
+        {
+            PyErr_Clear();
+            PyErr_SetString(PyExc_TypeError, "overflow error converting int to C long long");
+        }
+        return 0;
+    }
+    (*value) = val;
+    return 1;
+}
+#endif /* defined(ZODB_UNSIGNED_VALUE_INTS) || defined(ZODB_UNSIGNED_KEY_INTS) */
+
 static PyObject *
 longlong_as_object(PY_LONG_LONG val)
 {
@@ -107,10 +206,11 @@ longlong_as_object(PY_LONG_LONG val)
     return INT_FROM_LONG((long)val);
 }
 
-
 static int
 longlong_convert(PyObject *ob, PY_LONG_LONG *value)
 {
+    PY_LONG_LONG val;
+    int overflow;
 #ifndef PY3K
     if (PyInt_Check(ob))
     {
@@ -124,20 +224,15 @@ longlong_convert(PyObject *ob, PY_LONG_LONG *value)
         PyErr_SetString(PyExc_TypeError, "expected integer key");
         return 0;
     }
-    else
+    val = PyLong_AsLongLongAndOverflow(ob, &overflow);
+    if (!longlong_handle_overflow(val, overflow))
     {
-        PY_LONG_LONG val;
-        int overflow;
-        val = PyLong_AsLongLongAndOverflow(ob, &overflow);
-        if (overflow)
-            goto overflow;
-        (*value) = val;
-        return 1;
+        return 0;
     }
-overflow:
-    PyErr_SetString(PyExc_ValueError, "long integer out of range");
-    return 0;
+    (*value) = val;
+    return 1;
 }
+
 #endif  /* NEED_LONG_LONG_SUPPORT */
 
 
@@ -220,6 +315,7 @@ typedef struct BTree_s {
   long max_leaf_size;
 } BTree;
 
+static PyTypeObject BTreeTypeType;
 static PyTypeObject BTreeType;
 static PyTypeObject BucketType;
 
@@ -327,8 +423,8 @@ IndexError(int i)
 
     v = INT_FROM_LONG(i);
     if (!v) {
-	v = Py_None;
-	Py_INCREF(v);
+        v = Py_None;
+        Py_INCREF(v);
     }
     PyErr_SetObject(PyExc_IndexError, v);
     Py_DECREF(v);
@@ -355,20 +451,20 @@ PreviousBucket(Bucket **current, Bucket *first)
 
     do {
         trailing = first;
-	PER_USE_OR_RETURN(first, -1);
+        PER_USE_OR_RETURN(first, -1);
         first = first->next;
 
-    	((trailing)->state==cPersistent_STICKY_STATE
-     	&&
-     	((trailing)->state=cPersistent_UPTODATE_STATE));
+        ((trailing)->state==cPersistent_STICKY_STATE
+        &&
+        ((trailing)->state=cPersistent_UPTODATE_STATE));
 
-    	PER_ACCESSED(trailing);
+        PER_ACCESSED(trailing);
 
-	if (first == *current) {
-	    *current = trailing;
-	    result = 1;
-	    break;
-	}
+        if (first == *current) {
+            *current = trailing;
+            result = 1;
+            break;
+        }
     } while (first);
 
     return result;
@@ -383,7 +479,7 @@ BTree_Malloc(size_t sz)
 
     r = malloc(sz);
     if (r)
-	return r;
+        return r;
 
     PyErr_NoMemory();
     return NULL;
@@ -397,12 +493,12 @@ BTree_Realloc(void *p, size_t sz)
     ASSERT(sz > 0, "non-positive size realloc", NULL);
 
     if (p)
-	r = realloc(p, sz);
+        r = realloc(p, sz);
     else
-	r = malloc(sz);
+        r = malloc(sz);
 
     UNLESS (r)
-	PyErr_NoMemory();
+        PyErr_NoMemory();
 
     return r;
 }
@@ -411,8 +507,25 @@ BTree_Realloc(void *p, size_t sz)
  * (iter)?(keys|values|items)
  */
 static char *search_keywords[] = {"min", "max",
-				  "excludemin", "excludemax",
-				  0};
+                                  "excludemin", "excludemax",
+                                  0};
+
+/**
+ * Call this instead of using `PyErr_ExceptionMatches(PyExc_KeyError)`
+ * when you intend to suppress the KeyError.
+ *
+ * We want to match only exactly ``PyExc_KeyError``, and not subclasses
+ * such as ``ZODB.POSException.POSKeyError``.
+ */
+static int
+BTree_ShouldSuppressKeyError()
+{
+    PyObject* exc_type = PyErr_Occurred(); /* Returns borrowed reference. */
+    if (exc_type && exc_type == PyExc_KeyError) {
+        return 1;
+    }
+    return 0;
+}
 
 #include "BTreeItemsTemplate.c"
 #include "BucketTemplate.c"
@@ -423,37 +536,37 @@ static char *search_keywords[] = {"min", "max",
 #include "MergeTemplate.c"
 
 static struct PyMethodDef module_methods[] = {
-  {"difference", (PyCFunction) difference_m,	METH_VARARGS,
-   "difference(o1, o2) -- "
+  {"difference", (PyCFunction) difference_m,    METH_VARARGS,
+   "difference(o1, o2)\n"
    "compute the difference between o1 and o2"
   },
-  {"union", (PyCFunction) union_m,	METH_VARARGS,
-   "union(o1, o2) -- compute the union of o1 and o2\n"
+  {"union", (PyCFunction) union_m,      METH_VARARGS,
+   "union(o1, o2)\ncompute the union of o1 and o2\n"
   },
-  {"intersection", (PyCFunction) intersection_m,	METH_VARARGS,
-   "intersection(o1, o2) -- "
+  {"intersection", (PyCFunction) intersection_m,        METH_VARARGS,
+   "intersection(o1, o2)\n"
    "compute the intersection of o1 and o2"
   },
 #ifdef MERGE
-  {"weightedUnion", (PyCFunction) wunion_m,	METH_VARARGS,
-   "weightedUnion(o1, o2 [, w1, w2]) -- compute the union of o1 and o2\n"
+  {"weightedUnion", (PyCFunction) wunion_m,     METH_VARARGS,
+   "weightedUnion(o1, o2 [, w1, w2])\ncompute the union of o1 and o2\n"
    "\nw1 and w2 are weights."
   },
-  {"weightedIntersection", (PyCFunction) wintersection_m,	METH_VARARGS,
-   "weightedIntersection(o1, o2 [, w1, w2]) -- "
+  {"weightedIntersection", (PyCFunction) wintersection_m,       METH_VARARGS,
+   "weightedIntersection(o1, o2 [, w1, w2])\n"
    "compute the intersection of o1 and o2\n"
    "\nw1 and w2 are weights."
   },
 #endif
 #ifdef MULTI_INT_UNION
   {"multiunion", (PyCFunction) multiunion_m, METH_VARARGS,
-   "multiunion(seq) -- compute union of a sequence of integer sets.\n"
+   "multiunion(seq)\ncompute union of a sequence of integer sets.\n"
    "\n"
    "Each element of seq must be an integer set, or convertible to one\n"
    "via the set iteration protocol.  The union returned is an IISet."
   },
 #endif
-  {NULL,		NULL}		/* sentinel */
+  {NULL,                NULL}           /* sentinel */
 };
 
 static char BTree_module_documentation[] =
@@ -472,19 +585,50 @@ VALUEMACROS_H
 BTREEITEMSTEMPLATE_C
 ;
 
-int
-init_persist_type(PyTypeObject *type)
+static int
+init_type_with_meta_base(PyTypeObject *type, PyTypeObject* meta, PyTypeObject* base)
 {
+    int result;
+    PyObject* slotnames;
 #ifdef PY3K
-    ((PyObject*)type)->ob_type = &PyType_Type;
+    ((PyObject*)type)->ob_type = meta;
 #else
-    type->ob_type = &PyType_Type;
+    type->ob_type = meta;
 #endif
-    type->tp_base = cPersistenceCAPI->pertype;
+    type->tp_base = base;
 
     if (PyType_Ready(type) < 0)
-	return 0;
+        return 0;
+    /*
+      persistent looks for __slotnames__ in the dict at deactivation time,
+      and if it's not present, calls ``copyreg._slotnames``, which itself
+      looks in the dict again. Then it does some computation, and tries to
+      store the object in the dict --- which for built-in types, it can't.
+      So we can save some runtime if we store an empty slotnames for these classes.
+    */
+    slotnames = PyTuple_New(0);
+    if (!slotnames) {
+        return 0;
+    }
+    result = PyDict_SetItem(type->tp_dict, __slotnames__str, slotnames);
+    Py_DECREF(slotnames);
+    return result < 0 ? 0 : 1;
+}
 
+int /* why isn't this static? */
+init_persist_type(PyTypeObject* type)
+{
+    return init_type_with_meta_base(type, &PyType_Type, cPersistenceCAPI->pertype);
+}
+
+static int init_tree_type(PyTypeObject* type, PyTypeObject* bucket_type)
+{
+    if (!init_type_with_meta_base(type, &BTreeTypeType, cPersistenceCAPI->pertype)) {
+        return 0;
+    }
+    if (PyDict_SetItem(type->tp_dict, _bucket_type_str, (PyObject*)bucket_type) < 0) {
+        return 0;
+    }
     return 1;
 }
 
@@ -533,6 +677,24 @@ module_init(void)
     max_leaf_size_str = INTERN("max_leaf_size");
     if (! max_leaf_size_str)
         return NULL;
+    __slotnames__str = INTERN("__slotnames__");
+    if (!__slotnames__str)
+        return NULL;
+
+    BTreeType_setattro_allowed_names = PyTuple_Pack(
+        5,
+        /* BTree attributes  */
+        max_internal_size_str,
+        max_leaf_size_str,
+        /* zope.interface attributes */
+        /*
+          Technically, INTERNING directly here leaks references,
+          but since we can't be unloaded, it's not a problem.
+        */
+        INTERN("__implemented__"),
+        INTERN("__providedBy__"),
+        INTERN("__provides__")
+    );
 
     /* Grab the ConflictError class */
     interfaces = PyImport_ImportModule("BTrees.Interfaces");
@@ -558,8 +720,18 @@ module_init(void)
     cPersistenceCAPI = (cPersistenceCAPIstruct *)PyCObject_Import(
                 "persistent.cPersistence", "CAPI");
 #endif
-    if (cPersistenceCAPI == NULL)
+    if (cPersistenceCAPI == NULL) {
+       /* The Capsule API attempts to import 'persistent' and then
+        * walk down to the specified attribute using getattr. If the C
+        * extensions aren't available, this can result in an
+        * AttributeError being raised. Let that percolate up as an
+        * ImportError so it can be caught in the expected way.
+        */
+       if (PyErr_Occurred() && !PyErr_ExceptionMatches(PyExc_ImportError)) {
+           PyErr_SetString(PyExc_ImportError, "persistent C extension unavailable");
+       }
         return NULL;
+   }
 
 #ifdef PY3K
 #define _SET_TYPE(typ) ((PyObject*)(&typ))->ob_type = &PyType_Type
@@ -573,25 +745,22 @@ module_init(void)
     SetType.tp_new = PyType_GenericNew;
     BTreeType.tp_new = PyType_GenericNew;
     TreeSetType.tp_new = PyType_GenericNew;
-    if (!init_persist_type(&BucketType))
-	    return NULL;
-    if (!init_persist_type(&BTreeType))
-	    return NULL;
-    if (!init_persist_type(&SetType))
-	    return NULL;
-    if (!init_persist_type(&TreeSetType))
-	    return NULL;
 
-    if (PyDict_SetItem(BTreeType.tp_dict, _bucket_type_str,
-		       (PyObject *)&BucketType) < 0)
-    {
-        fprintf(stderr, "btree failed\n");
+    if (!init_persist_type(&BucketType))
+            return NULL;
+
+    if (!init_type_with_meta_base(&BTreeTypeType, &PyType_Type, &PyType_Type)) {
         return NULL;
     }
-    if (PyDict_SetItem(TreeSetType.tp_dict, _bucket_type_str,
-		       (PyObject *)&SetType) < 0)
-    {
-        fprintf(stderr, "bucket failed\n");
+
+    if (!init_tree_type(&BTreeType, &BucketType)) {
+        return NULL;
+    }
+
+    if (!init_persist_type(&SetType))
+        return NULL;
+
+    if (!init_tree_type(&TreeSetType, &SetType)) {
         return NULL;
     }
 
@@ -600,43 +769,47 @@ module_init(void)
     module = PyModule_Create(&moduledef);
 #else
     module = Py_InitModule4("_" MOD_NAME_PREFIX "BTree",
-		       module_methods, BTree_module_documentation,
-		       (PyObject *)NULL, PYTHON_API_VERSION);
+                       module_methods, BTree_module_documentation,
+                       (PyObject *)NULL, PYTHON_API_VERSION);
 #endif
 
     /* Add some symbolic constants to the module */
     mod_dict = PyModule_GetDict(module);
+
     if (PyDict_SetItemString(mod_dict, MOD_NAME_PREFIX "Bucket",
-			     (PyObject *)&BucketType) < 0)
+                             (PyObject *)&BucketType) < 0)
         return NULL;
     if (PyDict_SetItemString(mod_dict, MOD_NAME_PREFIX "BTree",
-			     (PyObject *)&BTreeType) < 0)
+                             (PyObject *)&BTreeType) < 0)
         return NULL;
     if (PyDict_SetItemString(mod_dict, MOD_NAME_PREFIX "Set",
-			     (PyObject *)&SetType) < 0)
+                             (PyObject *)&SetType) < 0)
         return NULL;
     if (PyDict_SetItemString(mod_dict, MOD_NAME_PREFIX "TreeSet",
-			     (PyObject *)&TreeSetType) < 0)
+                             (PyObject *)&TreeSetType) < 0)
         return NULL;
     if (PyDict_SetItemString(mod_dict, MOD_NAME_PREFIX "TreeIterator",
-			     (PyObject *)&BTreeIter_Type) < 0)
+                             (PyObject *)&BTreeIter_Type) < 0)
         return NULL;
-	/* We also want to be able to access these constants without the prefix
-	 * so that code can more easily exchange modules (particularly the integer
-	 * and long modules, but also others).  The TreeIterator is only internal,
-	 * so we don't bother to expose that.
+        /* We also want to be able to access these constants without the prefix
+         * so that code can more easily exchange modules (particularly the integer
+         * and long modules, but also others).  The TreeIterator is only internal,
+         * so we don't bother to expose that.
      */
     if (PyDict_SetItemString(mod_dict, "Bucket",
-			     (PyObject *)&BucketType) < 0)
+                             (PyObject *)&BucketType) < 0)
         return NULL;
     if (PyDict_SetItemString(mod_dict, "BTree",
-			     (PyObject *)&BTreeType) < 0)
+                             (PyObject *)&BTreeType) < 0)
         return NULL;
     if (PyDict_SetItemString(mod_dict, "Set",
-			     (PyObject *)&SetType) < 0)
+                             (PyObject *)&SetType) < 0)
         return NULL;
     if (PyDict_SetItemString(mod_dict, "TreeSet",
-			     (PyObject *)&TreeSetType) < 0)
+                             (PyObject *)&TreeSetType) < 0)
+        return NULL;
+    if (PyDict_SetItemString(mod_dict, "TreeItems",
+                             (PyObject *)&BTreeItemsType) < 0)
         return NULL;
 #if defined(ZODB_64BIT_INTS) && defined(NEED_LONG_LONG_SUPPORT)
     if (PyDict_SetItemString(mod_dict, "using64bits", Py_True) < 0)
